
/**************************************************************
 *
 *		Functions for Two Wire Interface ( I2C )
 *
 **************************************************************/

#include "io.h"
#include "uart.h"
#include "types.h"
#include "config.h"

#include "twi.h"

#include "signal.h"
#include "interrupt.h"


static UCHAR TWI_buf[ TWI_BUFFER_SIZE ]; // Transceiver buffer
static UCHAR TWI_msgSize;                // Number of bytes to be transmitted.
static UCHAR TWI_state = TWI_NO_STATE;   // State byte. Default set to TWI_NO_STATE.


// TWI_statusReg is defined in twi.h
union TWI_statusReg TWI_statusReg = {0};            


/****************************************************************************
 *
 *		Initialize Two Wire Interface ( I2C )
 *
 ****************************************************************************/
void TWI_Master_Init(void)
{
	UINT16 TWBR_Val; 

	// Calculate Value for Bitrate Register without Prescaler 
	TWBR_Val = ( (F_CPU/TWI_CLOCK) - 16 ) / 2;

	if ( TWBR_Val <= 255 )
	{
		// no Prescaler needed
		TWSR =  0x00;
	} 
		else
		{
			// set Prescaler to 4
			TWSR =  (0<<TWPS1) | (1<<TWPS0); 

			// Re-Calculate Value for Bitrate Register with Prescaler
			TWBR_Val = ( (F_CPU/TWI_CLOCK) - 16 ) / ( 2 * 4 );


			// if Bitrate Register Value is still too big
			if ( TWBR_Val > 255 )
			{
				// set Prescaler to 16
				TWSR = (1<<TWPS1) | (0<<TWPS0);

				// Re-Calculate Value for Bitrate Register with larger Prescaler
				TWBR_Val = ( (F_CPU/TWI_CLOCK) - 16 ) / ( 2 * 16 );
			}
		}

	// load Bitrate Register with calculated Value
	TWBR = TWBR_Val;

				   
  	TWDR = 	0xFF;                             // Default content = SDA released.
  	TWCR = 	(1<<TWEN)|                        // Enable TWI-interface and release TWI pins.
         	(0<<TWIE)|(0<<TWINT)|             // Disable Interupt.
         	(0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|  // No Signal requests.
         	(0<<TWWC);                        //
}  
// end TWIMasterInit



/****************************************************************************
 *
 *	Check if the TWI_ISR is busy transmitting.
 *
 ****************************************************************************/
UCHAR TWI_Transceiver_Busy( void )
{
	// IF TWI Interrupt is enabled then the Transceiver is busy
	return ( TWCR & (1<<TWIE) );   
}
// end TWI_Transceiver_Busy


/****************************************************************************
 * Call this function to fetch the state information of the previous operation.
 * The function will hold execution (loop)
 * until the TWI_ISR has completed with the previous operation. 
 * If there was an error, then the function will return the TWI State code. 
 ***************************************************************************/
UCHAR TWI_Get_State_Info( void )
{
  while ( TWI_Transceiver_Busy() ); // Wait until TWI has completed the transmission.
  return ( TWI_state );             // Return error state.
}
// end TWI_Get_State_Info


/****************************************************************************
 * Call this function to send a prepared message. The first byte must contain
 * the slave address and the read/write bit. Consecutive bytes contain the
 * data to be sent, or empty locations for data to be read from the slave.
 * Also include how many bytes that should be sent/read including the 
 * address byte.
 * The function will hold execution (loop) until the TWI_ISR has completed
 * with the previous operation, then initialize the next operation and return.
 ****************************************************************************/
void TWI_Start_Transceiver_With_Data( unsigned char *msg, unsigned char msgSize )
{
	unsigned char temp;

	while ( TWI_Transceiver_Busy() );           // Wait until TWI is ready for next transmission.

	// Number of data to transmit
	TWI_msgSize = msgSize;       
	 
	// Store slave address with R/W setting               
	TWI_buf[0]  = msg[0]; 
	
	// If it is a write operation, then also copy data.                      
	if (!( msg[0] & ( TRUE <<TWI_READ_BIT) ))   
	{
    	for ( temp = 1; temp < msgSize; temp++ )
      	TWI_buf[ temp ] = msg[ temp ];
  	}

  	TWI_statusReg.all = 0;    
	  
  	TWI_state         = TWI_NO_STATE ;

  	TWCR = 	(1<<TWEN)|                           // TWI Interface enabled.
         	(1<<TWIE)|(1<<TWINT)|                // Enable TWI Interupt and clear the flag.
         	(0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|     // Initiate a START condition.
         	(0<<TWWC);                           //
}
// end TWI_Start_Transceiver_With_Data


/****************************************************************************
 * Call this function to resend the last message. The driver will reuse the
 * data previously put in the transceiver buffers.
 * The function will hold execution (loop) until the TWI_ISR has completed
 * with the previous operation,then initialize the next operation and return.
 ****************************************************************************/
void TWI_Start_Transceiver( void )
{
	// Wait until TWI is ready for next transmission.
  	while ( TWI_Transceiver_Busy() );             
  
  	TWI_statusReg.all = 0;      

  	TWI_state         = TWI_NO_STATE ;
  	TWCR = 	(1<<TWEN)|                       // TWI Interface enabled.
         	(1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag.
         	(0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)| // Initiate a START condition.
         	(0<<TWWC);                       //
}


/****************************************************************************
 * Call this function to read out the requested data from the TWI transceiver
 * buffer. I.e. first call TWI_Start_Transceiver to send a request for data
 * to the slave. Then Run this function to collect the data when they have
 * arrived. Include a pointer to where to place the data and the number of
 * bytes requested (including the address field) in the function call.
 * The function will hold execution (loop) until the TWI_ISR has completed
 * with the previous operation, before reading out the data and returning.
 * If there was an error in the previous transmission the function will
 * return the TWI error code.
 ****************************************************************************/
UCHAR TWI_Get_Data_From_Transceiver( unsigned char *msg, unsigned char msgSize )
{
  unsigned char i;

  while ( TWI_Transceiver_Busy() );             // Wait until TWI is ready for next transmission.

  if( TWI_statusReg.lastTransOK )               // Last transmission competed successfully.              
  {                                             
    for ( i=0; i<msgSize; i++ )                 // Copy data from Transceiver buffer.
    {
      msg[ i ] = TWI_buf[ i ];
    }
  }
  return( TWI_statusReg.lastTransOK );                                   
}
// end TWI_Get_Data_From_Transceiver


// ********** Interrupt Handlers ********** //
/****************************************************************************
 * This function is the Interrupt Service Routine (ISR), and called when the
 * TWI interrupt is triggered; that is whenever a TWI event has occurred.
 * This function should not be called directly from the main application.
 ****************************************************************************/
SIGNAL( SIG_2WIRE_SERIAL )
{
  static unsigned char TWI_bufPtr;
  
  switch (TWSR)
  {
    case TWI_START:             // START has been transmitted  
    case TWI_REP_START:         // Repeated START has been transmitted
      TWI_bufPtr = 0;                                     // Set buffer pointer to the TWI Address location
    case TWI_MTX_ADR_ACK:       // SLA+W has been tramsmitted and ACK received
    case TWI_MTX_DATA_ACK:      // Data byte has been tramsmitted and ACK received
      if (TWI_bufPtr < TWI_msgSize)
      {
        TWDR = TWI_buf[TWI_bufPtr++];
        TWCR = (1<<TWEN)|                                 // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|                      // Enable TWI Interupt and clear the flag to send byte
               (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|           //
               (0<<TWWC);                                 //  
      }else                    // Send STOP after last byte
      {
        TWI_statusReg.lastTransOK = TRUE;                 // Set status bits to completed successfully. 
        TWCR = (1<<TWEN)|                                 // TWI Interface enabled
               (0<<TWIE)|(1<<TWINT)|                      // Disable TWI Interrupt and clear the flag
               (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|           // Initiate a STOP condition.
               (0<<TWWC);                                 //
      }
      break;
    case TWI_MRX_DATA_ACK:      // Data byte has been received and ACK tramsmitted
      TWI_buf[TWI_bufPtr++] = TWDR;
    case TWI_MRX_ADR_ACK:       // SLA+R has been tramsmitted and ACK received
      if (TWI_bufPtr < (TWI_msgSize-1) )                  // Detect the last byte to NACK it.
      {
        TWCR = (1<<TWEN)|                                 // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|                      // Enable TWI Interupt and clear the flag to read next byte
               (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|           // Send ACK after reception
               (0<<TWWC);                                 //  
      }else                    // Send NACK after next reception
      {
        TWCR = (1<<TWEN)|                                 // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|                      // Enable TWI Interupt and clear the flag to read next byte
               (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|           // Send NACK after reception
               (0<<TWWC);                                 // 
      }    
      break; 
    case TWI_MRX_DATA_NACK:     // Data byte has been received and NACK tramsmitted
      TWI_buf[TWI_bufPtr] = TWDR;
      TWI_statusReg.lastTransOK = TRUE;                 // Set status bits to completed successfully. 
      TWCR = (1<<TWEN)|                                 // TWI Interface enabled
             (0<<TWIE)|(1<<TWINT)|                      // Disable TWI Interrupt and clear the flag
             (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|           // Initiate a STOP condition.
             (0<<TWWC);                                 //
      break;      
    case TWI_ARB_LOST:          // Arbitration lost
      TWCR = (1<<TWEN)|                                 // TWI Interface enabled
             (1<<TWIE)|(1<<TWINT)|                      // Enable TWI Interupt and clear the flag
             (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|           // Initiate a (RE)START condition.
             (0<<TWWC);                                 //
      break;
    case TWI_MTX_ADR_NACK:      // SLA+W has been tramsmitted and NACK received
    case TWI_MRX_ADR_NACK:      // SLA+R has been tramsmitted and NACK received    
    case TWI_MTX_DATA_NACK:     // Data byte has been tramsmitted and NACK received
//    case TWI_NO_STATE              // No relevant state information available; TWINT = 0
    case TWI_BUS_ERROR:         // Bus error due to an illegal START or STOP condition
    default:     
      TWI_state = TWSR;                                 // Store TWSR and automatically sets clears noErrors bit.
                                                        // Reset TWI Interface
      TWCR = (1<<TWEN)|                                 // Enable TWI-interface and release TWI pins
             (0<<TWIE)|(0<<TWINT)|                      // Disable Interupt
             (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|           // No Signal requests
             (0<<TWWC);                                 //
  }
}
// end TWI-ISR
